@@ -1,23 +1,22 @@
 from __future__ import annotations
 
+import datetime
+import json
 import re
 import traceback
 import typing as ty
 
 import discord
-import ast
-from ..auth import Record, Auth, AuthList
+
 from . import FluxCog
-from .. import utils
-import json
-from .. import FluxEvent, CommandEvent
+from .. import CommandEvent, utils
+from ..auth import Auth, AuthList, Record
 from ..command import Response
-from ..command.argh import Arg
+from ..context import GuildMessageCtx, ManualAuthCtx, ManualAuthorCtx
 from ..errors import CommandError
-from ..context import GuildMemberContext, ManualAuthContext
 
 if ty.TYPE_CHECKING:
-   from ..context import GuildMessageContext, AuthAwareContext, MessageContext
+   from ..context import CommandCtx, GuildAwareCtx
    from ..command import Command
 
 
@@ -31,9 +30,8 @@ class Builtins(FluxCog):
       "member"     : "member",
    }
 
-
    def load(self):
-      def parse_auth_id(ctx: GuildMessageContext, type_: str, target_: str) -> int:
+      def parse_auth_id(ctx: GuildAwareCtx, type_: str, target_: str) -> int:
          if type_ == "member":
             ids_ = utils.find_mentions(target_)
 
@@ -50,26 +48,34 @@ class Builtins(FluxCog):
                raise CommandError(f"No role found with ID {target_}")
             return role_id
          if type_ == "permissions":
-            p_dict = json.loads(target_)
-            return discord.Permissions(**p_dict).value
+            p_dict: ty.List[str] = json.loads(target_)
+            return discord.Permissions(**{p: True for p in p_dict}).value
 
-      def parse_auth_context(ctx: GuildMessageContext, type_: str, target_: str) -> ManualAuthContext:
+      def parse_auth_context(ctx: GuildAwareCtx, type_: str, target_: str) -> ManualAuthCtx:
          if type_ == "user":
             ids_ = utils.find_mentions(target_)
             if not ids_:
                raise CommandError(f"No user ID found in {target_}")
 
-            return ManualAuthContext(flux=self.flux, auth_list=AuthList(user=self.flux.get_user(ids_[0])), config_identifier=ids_[0])
+            return ManualAuthCtx(flux=self.flux, auth_list=AuthList(user=self.flux.get_user(ids_[0]).id), config_identifier=ids_[0])
          auth_id = parse_auth_id(ctx, type_=type_, target_=target_)
          if type_ == "member":
             member = ctx.guild.get_member(auth_id)
-            return ManualAuthContext(flux=self.flux, auth_list=AuthList(user=member, roles=[r.id for r in member.roles], permissions=member.guild_permissions),
-                                     config_identifier=str(ctx.guild.id))
+            if not member:
+               raise CommandError(f"No member found with id `{auth_id}`")
+            return ManualAuthCtx(flux=self.flux, auth_list=AuthList(user=member.id, roles=[r.id for r in member.roles], permissions=member.guild_permissions),
+                                 config_identifier=str(ctx.guild.id))
          if type_ == "role":
-            return ManualAuthContext(flux=self.flux, auth_list=AuthList(roles=[auth_id]), config_identifier=str(ctx.guild.id))
+            role = ctx.guild.get_role(auth_id)
+            if not role:
+               raise CommandError(f"No role found with id `{auth_id}`")
+            return ManualAuthCtx(flux=self.flux, auth_list=AuthList(roles=[auth_id], permissions=role.permissions), config_identifier=str(ctx.guild.id))
 
          if type_ == "permissions":
-            return ManualAuthContext(flux=self.flux, auth_list=AuthList(permissions=discord.Permissions(permissions=auth_id)), config_identifier=str(ctx.guild.id))
+            try:
+               return ManualAuthCtx(flux=self.flux, auth_list=AuthList(permissions=discord.Permissions(permissions=auth_id)), config_identifier=str(ctx.guild.id))
+            except TypeError as e:
+               raise CommandError(f"Permissions `{auth_id}` could not be parsed. See:\n{e}")
 
       # @CommandCheck.check(lambda ctx: ctx.author.id == self.flux.admin_id)
       # @self._commandeer(name="reload", parsed=False, private=True)
@@ -87,15 +93,16 @@ class Builtins(FluxCog):
       #    return Response()
 
       @self._commandeer(name="asif", parsed=False, default_auths=[Record.allow_all()])
-      async def __asif(ctx: GuildMessageContext, args: str):
+      async def __asif(ctx: CommandCtx, args: str, ):
          """
          asif [type] <target>/{target} command args*
          ==
          Runs `command` as if it was being run by a `target` user, member, role, or permission-haver
          ==
-         type: [user/role/member/permissions/u/r/m/p] the type of `target`;
-         target: <user/member/role>/{perms}. Simulates usage by a given user, member,
+         [type]: [user/role/member/permissions/u/r/m/p] the type of `target`;
+         <target>: <user/member/role>. Simulates usage by a given user, member,
          member with a role, or member that has a set of permissions. See s.ze.ax/perm for {perms} structure;
+         {target}: JSON Array of permissions to simulate having
          command: Name of the Command to run as `target`;
          args: command arguments to pass to the Command
          ==
@@ -104,7 +111,13 @@ class Builtins(FluxCog):
          :return:
          """
 
-         mock_type, mock_target, command = args.split(" ", 2)
+         try:
+            mock_type, mock_target, command, command_args = utils.regex_parse(re.compile(r"(\S+)\s+(\S+)\s+((\[[^\]]*\])|([^\s\[]+))\s+(\S+)\s*(.*)"),
+                                                                              args,
+                                                                              [x - 1 for x in [1, 2, 3, 6, 7]])
+         except (ValueError, AttributeError):
+            raise CommandError(f"See `help asif` for usage")
+
          MOCK_TYPES = {
             "u"   : "user",
             "user": "user",
@@ -116,26 +129,27 @@ class Builtins(FluxCog):
             raise CommandError(f"`{mock_type}` must be in [{', '.join(MOCK_TYPES.keys())}]")
 
          cmd_name, cmd_args, *_ = [*command.split(" ", 1), None]
-         auth_ctx = parse_auth_context(ctx=ctx, type_=mock_type, target_=mock_target)
+         mock_auth_ctx = parse_auth_context(ctx=ctx.msg_ctx, type_=mock_type, target_=mock_target)
+         mock_author_ctx = ManualAuthorCtx(ctx.msg_ctx.message.guild.get_member(mock_target)) if ctx.msg_ctx.message.guild else ctx.author_ctx
+
          cmd = utils.find_cmd_or_cog(self.flux, cmd_name, only="command")
+
          if not cmd:
             raise CommandError(f"Command {cmd_name} not found")
-         print("RUNNING ASIF WITH CTX")
-         print(ctx.config_identifier)
-         print(ctx)
-         if Auth.accepts(ctx, cmd):
-            await self.flux.router.submit(event=CommandEvent(flux=self.flux, msg_ctx=ctx, auth_ctx=auth_ctx, cmd_name=cmd_name, cmd_args=cmd_args))
+         if Auth.accepts_all(ctx.auth_ctxs + [mock_auth_ctx], cmd):
+            await self.flux.router.submit(
+               event=CommandEvent(flux=self.flux, cmd_ctx=CommandCtx(ctx.msg_ctx, mock_author_ctx, ctx.auth_ctxs + [mock_auth_ctx]), cmd_name=cmd_name, cmd_args=cmd_args))
          else:
-            raise CommandError(f"Can only mock commands you have access to")
+            raise CommandError(f"Can only mock commands that you have access to")
 
          return Response()
 
       @self._commandeer(name="setprefix", parsed=False, default_auths=[Record.allow_perm(discord.Permissions(manage_guild=True))])
-      async def __set_prefix(ctx: GuildMessageContext, prefix: str):
+      async def __set_prefix(ctx: CommandCtx, prefix: str):
          """
          setprefix prefix
          ==
-          Sets the bot prefix to `prefix`
+         Sets the bot prefix to `prefix`
          Ignores surrounding whitespace. Please don't.
          ==
          prefix: The string to put before a command name. Strips leading and trailing spaces.
@@ -144,12 +158,12 @@ class Builtins(FluxCog):
          :param prefix:
          :return:
          """
-         async with self.flux.CONFIG.writeable_conf(ctx) as cfg:
+         async with self.flux.CONFIG.writeable_conf(ctx.msg_ctx) as cfg:
             cfg["prefix"] = prefix.strip()
          return Response()
 
-      @self._commandeer(name="exec", parsed=False, default_auths=[Record.deny_all()])
-      async def __exec(ctx: GuildMessageContext, script: str):
+      @self._commandeer(name="exec", parsed=False, override_auths=[Record.deny_all()])
+      async def __exec(ctx: CommandCtx, script: str):
          """
          exec ute order 66
          ==
@@ -180,22 +194,30 @@ class Builtins(FluxCog):
                           f"```py\n{res}\n```"), trashable=True)
 
       @self._commandeer(name="auth", parsed=False, default_auths=[Record.allow_perm(discord.Permissions(manage_guild=True))])
-      async def __auth(ctx: GuildMessageContext, auth_str):
+      async def __auth(ctx: CommandCtx, auth_str):
          """
          auth name [rule] <id>/{perm} [id_type]
          ==
          Authorizes some group (member, has a role, or has a permission) to use a command or a cog
          ==
          name: Command name or Cog name;
-         rule: [ALLOW,DENY];
-         id: <member/role>/{perm} The target member or role or permission to allow;
-         id_type: [MEMBER, ROLE, PERMISSION] Whatever `id` corresponds to
+         [rule]: [ALLOW/DENY];
+         <id>: <member/role> The target member or role to allow;
+         {perm:} A permission JSON array representing a set of permissions that a user must have ALL of.
+         ex. ["manage_server","kick_members"]
+         id_type: [MEMBER/ROLE/PERMISSION] Whatever `id` corresponds to
          ==
          :param ctx:
          :param auth_str:
          :return:
          """
-         rule_subject, rule, rule_target_id_raw, rule_type = auth_str.split(" ")
+         try:
+            rule_subject, rule, rule_target_id_raw, rule_type = utils.regex_parse(re.compile(r"(\S+)\s+(\S+)\s+((\[[^\]]*\])|([^\s\[]+))\s+(\S+)"),
+                                                                                  auth_str,
+                                                                                  [x - 1 for x in [1, 2, 3, 6]])
+         except (ValueError, AttributeError):
+            raise CommandError(f"See `help auth` for usage")
+
          rule_subject = rule_subject.lower()
          rule = rule.upper()
          rule_type = rule_type.lower()
@@ -206,16 +228,81 @@ class Builtins(FluxCog):
          if rule not in ["ALLOW", "DENY"]:
             raise CommandError(f'rule {rule} not in ["ALLOW","DENY"]')
          try:
-            target_id = parse_auth_id(ctx, type_=self.RULETYPES[rule_type], target_=rule_target_id_raw)
+            target_id = parse_auth_id(ctx.msg_ctx, type_=self.RULETYPES[rule_type], target_=rule_target_id_raw)
          except KeyError:
             raise CommandError(f"Rule type {rule_type} not in {self.RULETYPES.keys()}")
 
          record = Record(rule=rule, target_id=target_id, target_type=rule_type.upper())
-         await Auth.add_record(ctx, auth_id=cmd_or_cog.auth_id, record=record)
+         await Auth.add_record(ctx.msg_ctx, auth_id=cmd_or_cog.auth_id, record=record)
          return Response(f"Added record {record}")
 
-      @self._commandeer(name="help", parsed=False, default_auths=[Record.allow_all()], provide_auth=True)
-      async def __get_help(ctx: GuildMessageContext, help_target: ty.Optional[str], *, auth_ctx: AuthAwareContext):
+      @self._commandeer(name="userinfo", parsed=False, default_auths=[Record.allow_perm(discord.Permissions(manage_guild=True))])
+      async def __userinfo(ctx: CommandCtx, target_raw):
+         """
+         userinfo (<user/member>)
+         ==
+         Authorizes some group (member, has a role, or has a permission) to use a command or a cog
+         ==
+         <user/member>: The target member/user to userinfo. Defaults to caller if not provided.
+         ==
+         :param ctx:
+         :param target_raw:
+         :return:
+         """
+         if not target_raw:
+            target = ctx.author_ctx.author
+         else:
+            if not (target := utils.find_mentions(target_raw)):
+               raise CommandError(f"Cannot find a user/member in `{target_raw}`. It should either be an ID or a mention")
+
+            if isinstance(ctx.msg_ctx, GuildMessageCtx):
+               target = ctx.msg_ctx.guild.get_member(target)
+            else:
+               target = await ctx.msg_ctx.flux.get_user(target)
+
+         embed = discord.Embed(title=f"{target}'s Userinfo",  color=target.color)
+         embed.set_thumbnail(url=str(target.avatar_url))
+         embed.add_field(name="Display Name", value=utils.copylink(target.display_name), inline=True)
+         embed.add_field(name="ID", value=utils.copylink(str(target.id)), inline=False)
+         embed.add_field(name="Latest Join", value=utils.copylink(target.joined_at.strftime("%I:%M:%S %p on %a, %b %d, %Y")), inline=False)
+         embed.add_field(name="Creation Date", value=utils.copylink(target.created_at.strftime("%I:%M:%S %p on %a, %b %d, %Y")), inline=False)
+
+         if isinstance(ctx.msg_ctx, GuildMessageCtx):
+            if target.color != discord.Color.default():
+               embed.add_field(name="Color", value=utils.copylink(str(target.color.value)), inline=False)
+            if target.premium_since:
+               delta = (datetime.datetime.utcnow() - target.premium_since).days
+               D_IN_M = 29.53
+               D_IN_Y = 365.25
+               if delta < 7:
+                  output = f"{delta} days"
+               elif delta < D_IN_M:  # average month length
+                  output = f"{delta // 7} weeks"
+               elif delta < D_IN_Y * 2 + 1:
+                  output = f"{delta // D_IN_M} months"
+               else:
+                  output = f"{delta // D_IN_Y} years"
+               embed.add_field(name="Boosting for ", value=output)
+            roles = ",".join(role.mention for role in (target.roles or ())[::-1])
+            if len(roles) >= 2048:
+               url = utils.haste(self.flux.aiohttp_session, "\n".join(f"{role.id}:{role.name}" for role in target.roles))
+               roles = f"(roles)[{url}]"
+            embed.add_field(name="Roles", value=roles)
+
+            embed.add_field(
+               name="Permissions in this channel",
+               value=f"[Permissions](https://discordapi.com/permissions.html#{target.permissions_in(ctx.msg_ctx.channel).value})",
+               inline=False
+            )
+            embed.add_field(
+               name="Server Permissions",
+               value=f"[Permissions](https://discordapi.com/permissions.html#{target.guild_permissions.value})",
+               inline=False
+            )
+         return Response(embed=embed)
+
+      @self._commandeer(name="help", parsed=False, default_auths=[Record.allow_all()])
+      async def __get_help(ctx: CommandCtx, help_target: ty.Optional[str]):
          """
          help (command_name)
          ==
@@ -229,9 +316,10 @@ class Builtins(FluxCog):
          :param auth_ctx: Auth Context
          :return:
          """
-         configs = self.flux.CONFIG.of(ctx)
+         print("HELP!")
+         configs = self.flux.CONFIG.of(ctx.msg_ctx)
          authorized_cmds: ty.Dict[str, Command] = {command.name: command for cog in self.flux.cogs for command in cog.commands if
-                                                   Auth.accepts(auth_ctx, command) and command.name != "help"}
+                                                   Auth.accepts_all(ctx.auth_ctxs, command) and command.name != "help"}
 
          if not help_target:
             help_embed = discord.Embed(title=f"{utils.EMOJIS['question']} Command Help", description=f"{configs['prefix']}help <command> for more info")
@@ -263,8 +351,6 @@ class Builtins(FluxCog):
 
          embed.add_field(name="Usage", value=cmd.short_usage, inline=False)
 
-         # print(cmd.long_usage)
-         print(cmd.param_usage)
          for arg, detail in cmd.param_usage:
             embed.add_field(name=arg.strip(), value=detail.strip(), inline=False)
             # embed.add_field(name=detail,value="", inline=False)
